import os
from dataclasses import dataclass
from typing import Optional

from ticket_cross_check.gitlab_models import GitlabIssue

GITLAB_BASE_URL = os.getenv('GITLAB_BASE_URL', 'https://gitlab.com')


@dataclass
class IssueFileMatch:
    iid: int
    filename: str
    line_nr: int
    matched_issue: Optional[GitlabIssue]

    def get_issue_link(self):
        """

        :return: https://gitlab.com/exb/engineering/example-project/-/issues/
        """
        return self.matched_issue.get_issue_link()

    def get_deep_link(self, branch: str, baseLink: str = None):
        """
        returns
        the main project link is extracted from the issue
        and main as branch
        :param prefix:
        :param branch:
        :return: https://gitlab.com/exb/engineering/example-project/-/blob/main/metrics.py#L3
        """
        if not baseLink:
            prefix = self.matched_issue.references.get_project_url()
        else:
            prefix = baseLink
        return f"{GITLAB_BASE_URL}/{prefix}/-/blob/{branch}/{self.filename}#{self.line_nr}"

    def __hash__(self):
        return self.iid * 10000 + self.line_nr



import os

import pytest

from ticket_cross_check.gitlab_connector import GitlabConnector
from ticket_cross_check.gitlab_models import GitlabIssue


def need_token(func):
    @pytest.mark.skipif(os.getenv('PRIVATE_TOKEN') is None, reason="Need gitlab private token")
    def myfunc(*args, **kwargs):
        return func(*args, **kwargs)

    return myfunc


@need_token
def test_init_from_env():
    gc = GitlabConnector(personal_api_token=os.getenv('PRIVATE_TOKEN'), project_id=os.getenv('PROJECT_ID'))
    assert isinstance(gc, GitlabConnector)


@need_token
def test_factory():
    gc = GitlabConnector.factory()
    assert isinstance(gc, GitlabConnector)


@need_token
def test_get_issues():
    gc = GitlabConnector.factory()
    result = gc.get_issues()
    assert result
    assert isinstance(result, set)
    assert isinstance(result.pop(), GitlabIssue)



import os

from pydantic import BaseModel


class References(BaseModel):
    short: str
    full: str

    def get_project_url(self):
        return self.full.split('#')[0]


class GitlabIssue(BaseModel):
    id: int
    iid: int
    title: str
    state: str
    labels: list[str]
    references: References

    GITLAB_BASE_URL = os.getenv('GITLAB_BASE_URL', 'https://gitlab.com')

    def __hash__(self):
        """
        makes this object hashable
        :return:
        """
        return self.id

    def get_issue_link(self):
        """

        :return: https://gitlab.com/foo/group/project/-/issues/
        """
        prefix = self.references.get_project_url()
        return f"{self.GITLAB_BASE_URL}/{prefix}/-/issues/{self.iid}"



import os
import sys
from typing import Any

import gitlab
import requests
from loguru import logger

from ticket_cross_check.gitlab_models import GitlabIssue


class GitlabConnector:
    gitlab_baseurl = "https://gitlab.com/api/"
    api_version = "v4"

    def __init__(self, personal_api_token: str, project_id: str):
        assert personal_api_token  # is needed
        assert project_id  # is needed
        self.__personal_api_token = personal_api_token  # protect the token internally
        self.project_id = project_id

        self.gitlab = gitlab.Gitlab(private_token=personal_api_token)

    @staticmethod
    def factory() -> "GitlabConnector":
        personal_api_token = os.environ['PRIVATE_TOKEN']
        project_id = os.getenv('PROJECT_ID')
        if not project_id:
            # if project is not defined, assume current project
            project_id = os.environ['CI_PROJECT_ID']
        return GitlabConnector(personal_api_token, project_id)

    def _get(self, url: str) -> Any:
        """
        Raw getter, retrieves and checks API responses
        Exits hard on 401
        Raises ImportError on type mismatch
        """
        if '?' in url:
            url = f'{url}&private_token={self.__personal_api_token}'
        else:
            url = f'{url}?private_token={self.__personal_api_token}'

        result = requests.get(self.get_project_url(url))
        json = result.json()
        if result.status_code == 401:
            logger.error('API TOKEN invalid. Got 401')
            sys.exit(1)
        if result.status_code == 404:
            logger.error(f'URL not found {url}')
            sys.exit(1)
        return json

    def get_project_url(self, postfix: str = "") -> str:
        return self.gitlab_baseurl + self.api_version + f"/projects/{self.project_id}{postfix}"

    def get_issues_raw(self) -> list[dict]:
        """Get raw (full json) issues from gitlab via API"""
        return self._get('/issues')

    def get_issues(self) -> set[GitlabIssue]:
        return self._convert_issues(self.get_issues_raw())

    @staticmethod
    def _convert_issues(issue_json: list[dict]) -> set[GitlabIssue]:
        result = set()
        for raw_req in issue_json:
            result.add(GitlabIssue(**raw_req))
        return result

    def get_project_info_raw(self) -> dict:
        """Get raw project info from API"""
        return self._get(f'/projects/{self.project_id}', dict)

    def get_project(self):  # -> gitlab.v4.objects.projects.Project:
        """Get GitlabProject info from API"""
        rpi = self.gitlab.projects.get(self.project_id)
        return rpi



import json
from pathlib import Path

from loguru import logger

from ticket_cross_check.gitlab_models import GitlabIssue
from .models import IssueFileMatch


def _get_unmatched_issues_from_files(issues_in_files: dict[list[IssueFileMatch]], issues: set[GitlabIssue]) \
        -> list[IssueFileMatch]:
    unmatched = []
    iids = [int(issue.iid) for issue in issues]
    for iid in issues_in_files.keys():
        logger.trace(iid)
        if iid not in iids:
            logger.debug(issues_in_files[iid])
            unmatched += issues_in_files[iid]
    return unmatched


def _get_un_and_solved_issued(issues_in_files: dict[list[IssueFileMatch]], issues: set[GitlabIssue]) \
        -> tuple[set[GitlabIssue], set[IssueFileMatch]]:
    """

    :param issues_in_files:
    :param issues:
    :return: unsolved_issue set, solved_issue set
    """
    unsolved = set()
    solved = set()

    for issue in issues:
        if issue.iid not in issues_in_files:
            unsolved.add(issue)
        else:
            for iif in issues_in_files[issue.iid]:
                iif.matched_issue = issue
                solved.add(iif)
    return unsolved, solved


def analyze_issues_vs_code(issues_in_files: dict[list[IssueFileMatch]], issues: set[GitlabIssue]):
    # Unsolved Requirement: issues not in files
    # Problem: file-issues without real issues
    # Solved: issue in file(s)
    unsolved, solved = _get_un_and_solved_issued(issues_in_files, issues)
    problems = _get_unmatched_issues_from_files(issues_in_files, issues)
    return problems, unsolved, solved


def _write_json2file(path: Path, data: dict):
    with path.open(mode='w') as ofile:
        ofile.write(json.dumps(data, indent=2))


def render(
        solved: set[IssueFileMatch],
        unsolved: set[GitlabIssue],
        problems: list[IssueFileMatch],
        base_link: str,
        search_dir: str = ""
):
    data = []
    for ifm in solved:
        data.append([
            ifm.iid,
            f"<a href='{ifm.get_issue_link()}'>{ifm.matched_issue.title}</a>",
            f"<a href='{ifm.get_deep_link('main')}'>{ifm.filename}</a>",
            ifm.line_nr,
            "solved",
            search_dir
        ])

    status = "unsolved"
    for issue in unsolved:
        data.append([
            issue.iid,
            f"<a href='{issue.get_issue_link()}'>{issue.title}</a>",
            '',
            '',
            status,
            search_dir
        ])

    status = "problem"
    _id = "<span style='color: red'>%s</a>"
    for ifm in problems:
        data.append([
            _id % ifm.iid,
            '',
            f"<a href='{ifm.get_deep_link('main', base_link)}'>{ifm.filename}</a>",
            ifm.line_nr,
            status,
            search_dir
        ])

    return data
    # result = {
    #     'data': data,
    # }
    # _write_json2file(path, result)


def generate_base_matrix(gitlab_issues: set[GitlabIssue], dirs: list):
    matrix = {'_subdirs': dirs}

    for issue in gitlab_issues:
        matrix[issue.iid] = {'issue': issue, 'iid': issue.iid, 'counts': [0] * len(dirs)}
    return matrix


def build_issue_matrix(
        matrix_in: dict,
        solved: set[IssueFileMatch],
        search_dir: str = ""
):
    """
    returns {1: {'counts': [<count for first dir>, <count for second dir>],
     'iid': 1,
     'issue': GitlabIssue(id=112351387, iid=1, title='get issues from gitlab via api and personal token',
     state='closed', labels=[], references=References(short='#1', full='exb/engineering/ticket-cross-check#1'),
      GITLAB_BASE_URL='https://gitlab.com')},

    :param matrix_in:
    :param solved:
    :param search_dir:
    :return:
    """
    matrix = matrix_in.copy()
    count_index = matrix['_subdirs'].index(search_dir)
    for ifm in solved:
        issue_id = ifm.matched_issue.iid
        matrix[issue_id]['counts'][count_index] += 1
    return matrix


def render_matrix(matrix: dict):
    data = []
    issuelink = None
    for issue, value in matrix.items():
        if isinstance(issue, str):
            continue
        if not issuelink:
            end = value['issue'].get_issue_link().rfind('/')
            issuelink = value['issue'].get_issue_link()[:end]
        data.append(
            [
                issue,
                value['issue'].title,
            ] + value['counts']
        )

    return {
        "title": matrix['_subdirs'],
        "issuelink": issuelink,
        "data": data
    }



"""
Ticket Cross Checker
Find tickets (issues) in files and match them against the issue tracker (gitlab)
"""
import argparse
import os.path
import shutil
import sys
from pathlib import Path

from loguru import logger

from ticket_cross_check.gitlab_connector import GitlabConnector
from ticket_cross_check.issue_finder import scan_files
from ticket_cross_check.matcher import analyze_issues_vs_code, render, _write_json2file, generate_base_matrix, \
    build_issue_matrix, render_matrix


@logger.catch
def discover():
    parser = argparse.ArgumentParser()
    parser.add_argument('dirs', metavar='dir', type=str, nargs='+',
                        help='Sapce separated list of directories to process')
    parser.add_argument('-q', '--quiet', action='store_true', help="Show only warnigns and errors")
    parser.add_argument('-d', '--debug', action='store_true', help="Show more context")
    parser.add_argument('-t', '--trace', action='store_true', help="Show even more context / trace")
    parser.add_argument('-o', '--output', type=str, default='public', help="Where to write the output files")
    parser.add_argument('-e', '--exclude', type=str, default='', help="Comma separated list of dirnames to exclude")
    args = parser.parse_args()

    _setup_logger(args)

    file_issues = dict()
    for _dir in args.dirs:
        logger.info(f"--> Processing directory: {_dir} ({Path(_dir).absolute()})")
        file_issues[_dir] = scan_files(Path(_dir), exclude_dirs=args.exclude.split(','))
        logger.debug(f"<-- done with {_dir}")

    gitlab = GitlabConnector.factory()
    gitlab_issues = gitlab.get_issues()
    gitlab_project = gitlab.get_project()

    data = []
    issue_matrix = generate_base_matrix(gitlab_issues, args.dirs)
    for _dir in args.dirs:
        problems, unsolved, solved = analyze_issues_vs_code(file_issues[_dir], gitlab_issues)
        data += render(solved, unsolved, problems, base_link=gitlab_project.web_url, search_dir=str(_dir))
        issue_matrix = build_issue_matrix(issue_matrix, solved, search_dir=str(_dir))

    matrix_data = render_matrix(issue_matrix)

    output_dir = args.output
    os.makedirs(output_dir, exist_ok=True)
    full_json = Path(output_dir) / 'allinone.json'
    full_matrix = Path(output_dir) / 'matrix.json'
    # copy index template
    copy_resource('index.html', output_dir)
    copy_resource('matrix.html', output_dir)
    # write full json table
    _write_json2file(full_json, {"data": data})
    # write matrix
    _write_json2file(full_matrix, matrix_data)

    logger.info(f"Wrote all files to {Path(output_dir).absolute()}")


def copy_resource(filename, output_dir):
    index_from = Path(os.path.dirname(__file__)) / '..' / 'resource' / filename
    index_to = Path(output_dir).joinpath(filename).absolute()
    shutil.copy2(index_from, index_to)


def _setup_logger(args):
    level = "INFO"
    if args.trace:
        level = "TRACE"
    elif args.debug:
        level = "DEBUG"
    elif args.quiet:
        level = "WARN"
    logger.remove()
    logger.add(sys.stdout, colorize=True, level=level)
